"""
URLs for the forum app.
"""

# Django
from django.urls import path

# AA Forum
from aa_forum.constants import INTERNAL_URL_PREFIX
from aa_forum.views import forum

# Forum "public" URLs
urls = [
    path(route="", view=forum.index, name="forum_index"),
    # Forum URLs :: Board
    path(
        route="<slug:category_slug>/<slug:board_slug>/",
        view=forum.board,
        name="forum_board",
    ),
    path(
        route="<slug:category_slug>/<slug:board_slug>/page/<int:page_number>/",
        view=forum.board,
        name="forum_board",
    ),
    path(
        route=(
            f"{INTERNAL_URL_PREFIX}/<slug:category_slug>/<slug:board_slug>/new-topic/"
        ),
        view=forum.board_new_topic,
        name="forum_board_new_topic",
    ),
    # Forum URLs :: Topic
    path(
        route="<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/",
        view=forum.topic,
        name="forum_topic",
    ),
    path(
        route=(
            "<slug:category_slug>/<slug:board_slug>/"
            "<slug:topic_slug>/page/<int:page_number>/"
        ),
        view=forum.topic,
        name="forum_topic",
    ),
    path(
        route="<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/reply/",
        view=forum.topic_reply,
        name="forum_topic_reply",
    ),
    path(
        route="<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/unread/",
        view=forum.topic_first_unread_message,
        name="forum_topic_first_unread_message",
    ),
    path(
        route="<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/modify/",
        view=forum.topic_modify,
        name="forum_topic_modify",
    ),
    # Forum URLs :: Message
    path(
        route=(
            "<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/<int:message_id>/"
        ),
        view=forum.message,
        name="forum_message",
    ),
    path(
        route=(
            "<slug:category_slug>/<slug:board_slug>/"
            "<slug:topic_slug>/<int:message_id>/modify/"
        ),
        view=forum.message_modify,
        name="forum_message_modify",
    ),
]



"""
URL config for the AA-Forum app
"""

# Django
from django.urls import include, path

# AA Forum
from aa_forum.constants import INTERNAL_URL_PREFIX
from aa_forum.urls import forum, internal

app_name: str = "aa_forum"

# Put it all together
# IMPORTANT
# All internal URLs need to start with the designated prefix `{INTERNAL_URL_PREFIX}` to
# prevent conflicts with user-generated forum URLs.
urlpatterns = [
    # Forum internal URLs (Need to be first in line)
    path(f"{INTERNAL_URL_PREFIX}/", include(internal.urls)),
    # Forum "public" URLs
    path("", include(forum.urls)),
]



"""
Forms
"""

# Django
from django import forms
from django.contrib.auth.models import Group, User
from django.core.exceptions import ValidationError
from django.core.validators import MaxLengthValidator, URLValidator
from django.forms import ModelForm
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# CKEditor
from django_ckeditor_5.widgets import CKEditor5Widget

# AA Forum
from aa_forum.app_settings import discord_messaging_proxy_available
from aa_forum.helper.forms import message_empty
from aa_forum.helper.text import string_cleanup
from aa_forum.models import (
    Board,
    Category,
    General,
    Message,
    PersonalMessage,
    Setting,
    Topic,
    UserProfile,
)


def get_mandatory_form_label_text(text):
    """
    Returns a label text with a mandatory marker

    :param text:
    :type text:
    :return:
    :rtype:
    """

    required_text = _("This field is mandatory")
    required_marker = (
        f'<span aria-label="{required_text}" class="form-required-marker">*</span>'
    )

    return mark_safe(
        f'<span class="form-field-required">{text} {required_marker}</span>'
    )


class SpecialModelChoiceIterator(forms.models.ModelChoiceIterator):
    """
    Variant of Django's ModelChoiceIterator to prevent it from always re-fetching the
    given queryset from the database.
    """

    def __iter__(self):
        """
        Iterate over the choices

        :return:
        :rtype:
        """

        if self.field.empty_label is not None:
            yield "", self.field.empty_label

        queryset = self.queryset

        for obj in queryset:
            yield self.choice(obj=obj)


class SpecialModelMultipleChoiceField(forms.ModelMultipleChoiceField):
    """
    Variant of Django's ModelMultipleChoiceField to prevent it from always
    re-fetching the given queryset from the database.
    """

    iterator = SpecialModelChoiceIterator

    def _get_queryset(self):
        """
        Get the queryset

        :return:
        :rtype:
        """

        return self._queryset

    def _set_queryset(self, queryset):
        """
        Set the queryset

        :param queryset:
        :type queryset:
        :return:
        :rtype:
        """

        self._queryset = queryset
        self.widget.choices = self.choices

    queryset = property(_get_queryset, _set_queryset)


class NewTopicForm(ModelForm):
    """
    New topic form
    """

    # Non-model field, so we have to define it here and not in the Meta class.
    message = forms.CharField(
        widget=CKEditor5Widget(
            config_name="extends",
            attrs={
                "class": "aa-forum-ckeditor django_ckeditor_5",
                "rows": 10,
                "cols": 20,
                "style": "width: 100%;",
            },
        ),
        required=False,  # We have to set this to False, otherwise CKEditor5 will not work
        label=get_mandatory_form_label_text(text=_("Message")),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Topic

        fields = ["subject", "message"]
        labels = {"subject": get_mandatory_form_label_text(text=_("Subject"))}
        widgets = {
            "subject": forms.TextInput(attrs={"placeholder": _("Subject")}),
        }

    def clean(self):
        """
        Clean the form

        :return:
        :rtype:
        """

        cleaned_data = super().clean()

        if message_empty(message=cleaned_data.get("message")):
            raise ValidationError(_("You have forgotten the message!"))

        return cleaned_data

    def clean_message(self):
        """
        Cleanup the message

        :return:
        :rtype:
        """

        message = string_cleanup(string=self.cleaned_data["message"])

        if not message:
            raise ValidationError(_("You have forgotten the message!"))

        return message


class EditTopicForm(ModelForm):
    """
    Edit category form
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Topic

        fields = ["subject"]
        labels = {"subject": get_mandatory_form_label_text(text=_("Subject"))}
        widgets = {
            "subject": forms.TextInput(attrs={"placeholder": _("Subject")}),
        }


class NewCategoryForm(ModelForm):
    """
    New category form
    """

    # Non-model field, so we have to define it here and not in the Meta class.
    boards = forms.CharField(
        required=False,
        label=_("Boards"),
        help_text=_(
            "Boards to be created with this category (One board per line). These "
            "boards will have no group restrictions, so you have to add them later "
            "where needed."
        ),
        widget=forms.Textarea(
            attrs={
                "rows": 10,
                "cols": 20,
                "input_type": "textarea",
                "placeholder": _("Boards"),
            }
        ),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Category

        fields = ["name", "boards"]
        labels = {"name": get_mandatory_form_label_text(text=_("Name"))}
        widgets = {"name": forms.TextInput(attrs={"placeholder": _("Category name")})}


class EditCategoryForm(ModelForm):
    """
    Edit category form
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Category

        fields = ["name"]
        labels = {"name": get_mandatory_form_label_text(text=_("Name"))}
        widgets = {"name": forms.TextInput(attrs={"placeholder": _("Category name")})}


class EditBoardForm(ModelForm):
    """
    Edit board form
    """

    def __init__(self, *args, **kwargs):
        """
        When form is initialized

        :param args:
        :param kwargs:
        """

        groups_queryset = kwargs.pop("groups_queryset", None)

        super().__init__(*args, **kwargs)

        if groups_queryset:
            self.fields["groups"].queryset = groups_queryset

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Board

        fields = [
            "name",
            "description",
            "groups",
            "discord_webhook",
            "use_webhook_for_replies",
            "is_announcement_board",
            "announcement_groups",
        ]
        labels = {
            "name": get_mandatory_form_label_text(text=_("Name")),
            "description": _("Description"),
            "groups": _("Group restrictions"),
            "discord_webhook": _("Discord webhook (optional)"),
            "use_webhook_for_replies": _(
                "Use this Discord webhook for replies as well?"
            ),
            "is_announcement_board": _('Mark board as "Announcement Board"'),
            "announcement_groups": _(
                'Start topic restrictions for "Announcement Boards"'
            ),
        }
        widgets = {
            "name": forms.TextInput(attrs={"placeholder": _("Board name")}),
            "description": forms.Textarea(
                attrs={
                    "rows": 10,
                    "cols": 20,
                    "input_type": "textarea",
                    "placeholder": _("Board description (optional)"),
                }
            ),
            "discord_webhook": forms.TextInput(
                attrs={
                    "placeholder": "https://discord.com/api/webhooks/xxxxxxxxxxxxxxxxxx/xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"  # pylint: disable=line-too-long
                }
            ),
        }
        querysets = {
            "groups": Group.objects.none(),
            "announcement_groups": Group.objects.none(),
        }


class EditMessageForm(ModelForm):
    """
    Edit message form
    """

    # Non-model fields, so we have to define them here and not in the Meta class.
    close_topic = forms.BooleanField(
        required=False,
        label=_("Close topic"),
        help_text=_(
            "If checked, this topic will be closed after posting this message."
        ),
    )

    reopen_topic = forms.BooleanField(
        required=False,
        label=_("Reopen topic"),
        help_text=_(
            "If checked, this topic will be reopened after posting this message."
        ),
    )

    def __init__(self, *args, **kwargs):
        """
        When form is initialized

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        """

        super().__init__(*args, **kwargs)

        # We have to set this to False, otherwise CKEditor5 will not work
        self.fields["message"].required = False

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Message

        fields = ["message", "close_topic", "reopen_topic"]
        labels = {"message": get_mandatory_form_label_text(text=_("Message"))}

    def clean(self):
        """
        Clean the form

        :return:
        :rtype:
        """

        cleaned_data = super().clean()

        if message_empty(message=cleaned_data.get("message")):
            raise ValidationError(_("You have forgotten the message!"))

        return cleaned_data

    def clean_message(self):
        """
        Cleanup the message

        :return:
        :rtype:
        """

        message = string_cleanup(string=self.cleaned_data["message"])

        if not message:
            raise ValidationError(_("You have forgotten the message!"))

        return message


class UserProfileForm(ModelForm):
    """
    Userprofile form
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = UserProfile

        fields = [
            "signature",
            "website_title",
            "website_url",
            "discord_dm_on_new_personal_message",
            "show_unread_topics_dashboard_widget",
        ]
        help_texts = {
            "signature": _("Your signature will appear below your posts."),
            "website_title": _("Your website's title."),
            "website_url": _(
                "Your website's URL. (Don't forget to also set a title for your "
                "website, otherwise this field will be ignored.)"
            ),
            "discord_dm_on_new_personal_message": (
                _(
                    "Information: There is currently no module installed that can "
                    "handle Discord direct messages. Have a chat with your IT guys "
                    "to remedy this."
                )
                if discord_messaging_proxy_available() is False
                else ""
            ),
            "show_unread_topics_dashboard_widget": (
                _(
                    "Activating this setting will ad a widget to your dashboard that "
                    "shows unread topics in the forum."
                )
            ),
        }
        labels = {
            "signature": _("Signature"),
            "website_title": _("Website title"),
            "website_url": _("Website URL"),
            "discord_dm_on_new_personal_message": _(
                "PM me on Discord when I get a new personal message"
            ),
            "show_unread_topics_dashboard_widget": _(
                "Show unread topics as widget on the dashboard"
            ),
        }
        widgets = {
            "website_title": forms.TextInput(
                attrs={
                    "placeholder": _("e.g.: My Homepage"),
                }
            ),
            "website_url": forms.TextInput(
                attrs={
                    "placeholder": "https://example.com",
                }
            ),
        }

    def clean_signature(self):
        """
        Check that the signature is not longer than allowed

        :return:
        :rtype:
        """

        signature = string_cleanup(string=self.cleaned_data["signature"])

        if not signature:
            return ""

        max_signature_length = Setting.objects.get_setting(
            setting_key=Setting.Field.USERSIGNATURELENGTH
        )

        try:
            MaxLengthValidator(max_signature_length)(signature)
        except ValidationError as exc:
            raise ValidationError(
                _(
                    f"Ensure your signature has at most {max_signature_length} characters. (Currently: {len(signature)})"  # pylint: disable=line-too-long
                )
            ) from exc

        return signature

    def clean_website_url(self):
        """
        Check if it's a valid URL

        :return:
        :rtype:
        """

        website_url = self.cleaned_data["website_url"]

        if not website_url:
            return ""

        try:
            URLValidator()(website_url)
        except ValidationError as exc:
            raise ValidationError(_("This is not a valid URL")) from exc

        return website_url


class SettingForm(ModelForm):
    """
    Edit message form
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Setting

        fields = ["messages_per_page", "topics_per_page", "user_signature_length"]
        labels = {
            "messages_per_page": get_mandatory_form_label_text(
                Setting.Field.MESSAGESPERPAGE.label  # pylint: disable=no-member
            ),
            "topics_per_page": get_mandatory_form_label_text(
                Setting.Field.TOPICSPERPAGE.label  # pylint: disable=no-member
            ),
            "user_signature_length": get_mandatory_form_label_text(
                Setting.Field.USERSIGNATURELENGTH.label  # pylint: disable=no-member
            ),
        }


class NewPersonalMessageForm(ModelForm):
    """
    New personal message form
    """

    def __init__(self, *args, **kwargs):
        """
        When form is initialized

        :param args:
        :param kwargs:
        """

        super().__init__(*args, **kwargs)

        self.fields["recipient"].queryset = General.users_with_basic_access()
        self.fields["message"].required = False

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = PersonalMessage

        fields = ["recipient", "subject", "message"]
        labels = {
            "message": get_mandatory_form_label_text(text=_("Message")),
            "recipient": get_mandatory_form_label_text(text=_("Recipient")),
            "subject": get_mandatory_form_label_text(text=_("Subject")),
        }
        querysets = {"recipient": User.objects.none()}

    def clean(self):
        """
        Clean the form

        :return:
        :rtype:
        """

        cleaned_data = super().clean()

        if message_empty(message=cleaned_data.get("message")):
            raise ValidationError(_("You have forgotten the message!"))

        return cleaned_data

    def clean_message(self):
        """
        Cleanup the message

        :return:
        :rtype:
        """

        message = string_cleanup(string=self.cleaned_data["message"])

        if not message:
            raise ValidationError(_("You have forgotten the message!"))

        return message


class ReplyPersonalMessageForm(ModelForm):
    """
    Reply to personal message form
    """

    def __init__(self, *args, **kwargs):
        """
        When form is initialized

        :param args:
        :param kwargs:
        """

        super().__init__(*args, **kwargs)

        self.fields["message"].required = False

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = PersonalMessage

        fields = ["message"]
        labels = {"message": get_mandatory_form_label_text(text=_("Message"))}

    def clean(self):
        """
        Clean the form

        :return:
        :rtype:
        """

        cleaned_data = super().clean()

        if message_empty(message=cleaned_data.get("message")):
            raise ValidationError(_("You have forgotten the message!"))

        return cleaned_data

    def clean_message(self):
        """
        Cleanup the message

        :return:
        :rtype:
        """

        message = string_cleanup(string=self.cleaned_data["message"])

        if not message:
            raise ValidationError(_("You have forgotten the message!"))

        return message



"""
Test cases for the `aa_forum.helper.discord_messages` module `dhooks-lite` implementation.
"""

# Standard Library
from unittest import TestCase

# AA Forum
from aa_forum import __version__
from aa_forum.constants import APP_NAME_VERBOSE_USERAGENT, GITHUB_URL
from aa_forum.helper.discord_messages import _dhooks_lite_user_agent


class TestDhooksLiteUserAgent(TestCase):
    """
    Test cases for the `dhooks-lite` user agent
    """

    def test_create_useragent(self):
        """
        Test creating a user agent

        :return:
        :rtype:
        """

        obj = _dhooks_lite_user_agent()

        self.assertEqual(first=obj.name, second=APP_NAME_VERBOSE_USERAGENT)
        self.assertEqual(first=obj.url, second=GITHUB_URL)
        self.assertEqual(first=obj.version, second=__version__)

    def test_useragent_str(self):
        """
        Test the string representation of the user agent

        :return:
        :rtype:
        """

        obj = _dhooks_lite_user_agent()

        self.assertEqual(
            first=str(obj),
            second=f"{APP_NAME_VERBOSE_USERAGENT} ({GITHUB_URL}, {__version__})",
        )
